# -------------------------------------------------------------------------------------------------
# test_math2_mp_xi2.py: Calculating with Pools in Multiprocessing                              
#
# Theory:      Multiprocessingand simulataneous operations
#
# Description: Synchronization types like locks, conditions and queues are handy in multiprocessing
#
# class multiprocessing.Process([group[, target[, name[, args[, kwargs]]]]]) 
# 
#    Process objects represent activity that is run in a separate process. 
#    The Process class has equivalents of all the methods of threading.Thread.
#    
#constructor
#    should always be called with keyword arguments. 
#    
#group
#    should always be None; it exists solely for compatibility with Thread. 
#    
#target
#    is the callable object to be invoked by the run() method. 
#    It defaults to None, meaning nothing is called. 
#    
#name
#    is the process name
#    by default, a unique name is constructed of the form Process-N1:N2:...:Nk
#    where N1,N2,...,Nk is a sequence of integers whose length is determined by the generation of the process. 
#    
#args
#    is the argument tuple for the target invocation
#    
#kwargs
#    is a dictionary of keyword arguments for the target invocation. By default, no arguments are passed to target.
#    
#If a subclass overrides the constructor, it must make sure it invokes the base class constructor (Process.__init__()) before doing anything else to the process.
#    
#run() 
#    Method representing the process activity.
#    You may override this method in a subclass. 
#    The standard run() method invokes the callable object passed to the object constructor as the target argument, if any, 
#    with sequential and keyword arguments taken from the args and kwargs arguments, respectively.
#    
#start() 
#    Start the process activity.
#    This must be called at most once per process object. 
#    It arranges for the object run() method to be invoked in a separate process.
#    
#join([timeout]) 
#    Block the calling thread until the process whose join() method is called terminates or until the optional timeout occurs.
#    If timeout is None then there is no timeout.
#    A process can be joined many times.
#    A process cannot join itself because this would cause a deadlock. It is an error to attempt to join a process before it has been started.
#    
#name 
#    Return the process name.
#    
#Process.name = name 
#    Set the process name.
#    The name is a string used for identification purposes only. 
#    It has no semantics. 
#    Multiple processes may be given the same name. 
#    The initial name is set by the constructor.
#    
#is_alive() <--- OK
#    Return whether the process is alive.
#    Roughly, a process object is alive from the moment the start() method returns until the child process terminates.
#    
#daemon 
#    Return the process daemon flag., this is a boolean.
#    
#Process.daemon = daemonic 
#    Set the process daemon flag to the Boolean value daemonic. 
#    This must be called before start() is called.
#    The initial value is inherited from the creating process.
#    When a process exits, it attempts to terminate all of its daemonic child processes.
#    Note that a daemonic process is not allowed to create child processes. 
#    Otherwise a daemonic process would leave its children orphaned if it gets terminated when its parent process exits.
#        
#In addition process objects also support the following methods:
#    
#pid 
#    Return the process ID. Before the process is spawned, this will be None.
#    
#exitcode 
#    Return the child exit code. This will be None if the process has not yet terminated. A negative value -N indicates that the child was terminated by signal N.
#
#authkey 
#    Return the process authentication key (a byte string).
#    When multiprocessing is initialized the main process is assigned a random string using os.random().
#    When a Process object is created, it will inherit the authentication key of its parent process, although this may be changed using Process.authkey below.
#    See Authentication keys.
# 
#Process.authkey = authkey 
# Set the process authentication key which must be a byte string.
# 
#terminate() 
#    Terminate the process. 
#    On Unix this is done using the SIGTERM signal; 
#    on Windows TerminateProcess is used. Note that exit handlers and finally clauses, etc., will not be executed.
#    Note that descendant processes of the process will not be terminated, they will simply become orphaned
#    Warning
#    If this method is used when the associated process is using a pipe or queue 
#    then the pipe or queue is liable to become corrupted and may become unusable by other process. 
#    Similarly, if the process has acquired a lock or semaphore etc. then terminating it is liable to cause other processes to deadlock.
#    Note that the start(), join(), is_alive() and get_exit_code() methods should only be called by the process that created the process object.
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Imports from standard modules                              
# -------------------------------------------------------------------------------------------------
import os
import sys
from ctypes import Structure, c_double
import time
import threading
import gc
import random
import queue # <--- NOT Queue as in the original example
import operator


# -------------------------------------------------------------------------------------------------
# Imports from multiprocessing package                              
# -------------------------------------------------------------------------------------------------
import multiprocessing

from multiprocessing import Process
from multiprocessing import Queue
from multiprocessing import Pipe
from multiprocessing import Lock
from multiprocessing import Manager
from multiprocessing import Pool
from multiprocessing import current_process
from multiprocessing import freeze_support
from multiprocessing.sharedctypes import Value, Array
from multiprocessing.managers import BaseManager, BaseProxy

#from Queue import Empty <--- changed
from queue import Empty



# -------------------------------------------------------------------------------------------------
# yourDragonXi imports                               
# -------------------------------------------------------------------------------------------------
from process2xi import Process2Xi
from queue2xi  import Queue2Xi
from pipe2xi import Pipe2Xi
from lock2xi import Lock2Xi
from rlock2xi import RLock2Xi
from manager2xi import Manager2Xi 
from pool2xi import Pool2Xi
from value2xi import Value2Xi
from arrayfunction2xi import ArrayFunction2Xi #<--- note: this must be called (Array2Xi uses the class from managers.py)
from basemanager2xi import BaseManager2Xi
from baseproxy2xi import BaseProxy2Xi
from condition2xi import Condition2Xi  
from cpu_corecount2xi import cpu_corecount2xi
from current_process2xi import current_process2xi

# -------------------------------------------------------------------------------------------------
# Functions used by test code                              
# -------------------------------------------------------------------------------------------------
#
# -------------------------------------------------------------------------------------------------
# calculate
# -------------------------------------------------------------------------------------------------
def calculate(func, args):
    result = func(*args)
    return '%s says that %s%s = %s' % (
        multiprocessing.current_process().name,
        func.__name__, args, result
        )
# -------------------------------------------------------------------------------------------------
# calculatestar
# ------------------------------------------------------------------------------------------------
def calculatestar(args):
    return calculate(*args)
# -------------------------------------------------------------------------------------------------
# mul
# ------------------------------------------------------------------------------------------------
def mul(a, b):
    time.sleep(0.5*random.random())
    return a * b
# -------------------------------------------------------------------------------------------------
# plus
# ------------------------------------------------------------------------------------------------
def plus(a, b):
    time.sleep(0.5*random.random())
    return a + b
# -------------------------------------------------------------------------------------------------
# f
# ------------------------------------------------------------------------------------------------
def f(x):
    return 1.0 / (x-5.0)
# -------------------------------------------------------------------------------------------------
# pow3
# ------------------------------------------------------------------------------------------------
def pow3(x):
    return x**3
# -------------------------------------------------------------------------------------------------
# noop
# ------------------------------------------------------------------------------------------------
def noop(x):
    pass

    
# -------------------------------------------------------------------------------------------------
# calculate2xi function uses dragonized function current_process2xi in the module current_process2xi.py
# -------------------------------------------------------------------------------------------------
def calculate2xi(func, args):
    result = func(*args)
    return '%s says that %s%s = %s' % (
        current_process2xi().name,
        func.__name__, args, result
        )    
    
# -------------------------------------------------------------------------------------------------
# test1: 
# ------------------------------------------------------------------------------------------------
def test1():
    multiprocessing.freeze_support()    
    print ('cpu_count() = %d\n' % multiprocessing.cpu_count())

    #
    # Create pool
    #

    PROCESSES = 4
    print ('Creating pool with %d processes\n' % PROCESSES)
    pool = multiprocessing.Pool(PROCESSES)
    print ('pool = %s' % pool)
    
    #
    # Tests
    #

    TASKS = [(mul, (i, 7)) for i in range(10)] + \
            [(plus, (i, 8)) for i in range(10)]

    results = [pool.apply_async(calculate, t) for t in TASKS]
    imap_it = pool.imap(calculatestar, TASKS)
    imap_unordered_it = pool.imap_unordered(calculatestar, TASKS)

    print ('Ordered results using pool.apply_async():')
    for r in results:
        print ('\t', r.get())
    print()

    print ('Ordered results using pool.imap():')    
    for x in imap_it:
        print ('\t', x)
    print()

    print ('Unordered results using pool.imap_unordered():')
    for x in imap_unordered_it:
        print ('\t', x)
    print()

    print ('Ordered results using pool.map() --- will block till complete:')
    for x in pool.map(calculatestar, TASKS):
        print ('\t', x)
    print()

    #
    # Simple benchmarks
    #

    N = 100000
    print ('def pow3(x): return x**3')

    t = time.time()
    #A = map(pow3, range(N)) <--- range caused error
    A = map(pow3, range(N))
    print ('\tmap(pow3, range(%d)):\n\t\t%s seconds' % \
          (N, time.time() - t))

    t = time.time()
    B = pool.map(pow3, range(N))
    print ('\tpool.map(pow3, range(%d)):\n\t\t%s seconds' % \
          (N, time.time() - t))

    t = time.time()
    C = list(pool.imap(pow3, range(N), chunksize=N//8))
    print ('\tlist(pool.imap(pow3, range(%d), chunksize=%d)):\n\t\t%s' \
          ' seconds' % (N, N//8, time.time() - t))

    #assert A == B == C, (len(A), len(B), len(C)) <--- compilation error
    #assert (A == B == C, (len(A), len(B), len(C))) <--- unresolved, not even this ok !
    print ()
    L = [None] * 1000000
    print ('def noop(x): pass')
    print ('L = [None] * 1000000')

    t = time.time()
    A = map(noop, L)
    print ('\tmap(noop, L):\n\t\t%s seconds' % \
          (time.time() - t))

    t = time.time()
    B = pool.map(noop, L)
    print ('\tpool.map(noop, L):\n\t\t%s seconds' % \
          (time.time() - t))

    t = time.time()
    C = list(pool.imap(noop, L, chunksize=len(L)//8))
    print ('\tlist(pool.imap(noop, L, chunksize=%d)):\n\t\t%s seconds' % \
          (len(L)//8, time.time() - t))

    #assert (A == B == C, (len(A), len(B), len(C)))
    print ()
    del A, B, C, L

    #
    # Test error handling
    #

    print ('Testing error handling:')

    try:
        print (pool.apply(f, (5,)))
    except ZeroDivisionError:
        print ('\tGot ZeroDivisionError as expected from pool.apply()')
    else:
        #raise AssertionError, 'expected ZeroDivisionError'
        raise AssertionError('expected ZeroDivisionError')
    try:
        print (pool.map(f, range(10)))
    except ZeroDivisionError:
        print ('\tGot ZeroDivisionError as expected from pool.map()')
    else:
        raise AssertionError ('expected ZeroDivisionError')

    try:
        print (list(pool.imap(f, range(10))))
    except ZeroDivisionError:
        print ('\tGot ZeroDivisionError as expected from list(pool.imap())')
    else:
        raise AssertionError ('expected ZeroDivisionError')

    it = pool.imap(f, range(10))
    for i in range(10):
        try:
            x = it.next()
        except ZeroDivisionError:
            if i == 5:
                pass
        except StopIteration:
            break
        else:
            if i == 5:
                raise AssertionError ('expected ZeroDivisionError')

    assert (i == 9)
    print ('\tGot ZeroDivisionError as expected from IMapIterator.next()')
    print ()
    #
    # Testing timeouts
    #

    print ('Testing ApplyResult.get() with timeout:',)
    res = pool.apply_async(calculate, TASKS[0])
    while 1:
        sys.stdout.flush()
        try:
            sys.stdout.write('\n\t%s' % res.get(0.02))
            break
        except multiprocessing.TimeoutError:
            sys.stdout.write('.')
    print()


    print ('Testing IMapIterator.next() with timeout:')
    it = pool.imap(calculatestar, TASKS)
    while 1:
        sys.stdout.flush()
        try:
            sys.stdout.write('\n\t%s' % it.next(0.02))
        except StopIteration:
            break
        except multiprocessing.TimeoutError:
            sys.stdout.write('.')
    print()


    #
    # Testing callback
    #

    print ('Testing callback:')

    A = []
    B = [56, 0, 1, 8, 27, 64, 125, 216, 343, 512, 729]

    r = pool.apply_async(mul, (7, 8), callback=A.append)
    r.wait()

    r = pool.map_async(pow3, range(10), callback=A.extend)
    r.wait()

    if A == B:
        print ('\tcallbacks succeeded\n')
    else:
        print ('\t*** callbacks failed\n\t\t%s != %s\n' % (A, B))

    #
    # Check there are no outstanding tasks
    #

    # assert (not pool._cache, 'cache = %r' % pool._cache) <--- always true !?

    #
    # Check close() methods
    #

    print ('Testing close():')

    for worker in pool._pool:
        assert (worker.is_alive())

    result = pool.apply_async(time.sleep, [0.5])
    pool.close()
    pool.join()

    assert (result.get() is None)

    for worker in pool._pool:
        assert (not worker.is_alive())

    print ('\tclose() succeeded\n')

    #
    # Check terminate() method
    #

    print ('Testing terminate():')

    pool = multiprocessing.Pool(2)
    DELTA = 0.1
    ignore = pool.apply(pow3, [2])
    results = [pool.apply_async(time.sleep, [DELTA]) for i in range(100)]
    pool.terminate()
    pool.join()

    for worker in pool._pool:
        assert (not worker.is_alive())

    print ('\tterminate() succeeded\n')

    #
    # Check garbage collection
    #

    print ('Testing garbage collection:')

    pool = multiprocessing.Pool(2)
    DELTA = 0.1
    processes = pool._pool
    ignore = pool.apply(pow3, [2])
    results = [pool.apply_async(time.sleep, [DELTA]) for i in range(100)]

    results = pool = None

    time.sleep(DELTA * 2)

    for worker in processes:
        assert (not worker.is_alive())

    print ('\tgarbage collection succeeded\n')

# -------------------------------------------------------------------------------------------------
# test2: testing calculations with dragonized multiprocessing
# ------------------------------------------------------------------------------------------------
def test2():

    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test(00) cpu_corecount2xi() = %d\n' % cpu_corecount2xi()) 

    #
    # Create pool2xi
    #

    PROCESSES = 4
    print ('Creating dragonized pool with %d processes\n' % PROCESSES)
    pool2xi = Pool2Xi(PROCESSES)
    print ('pool2xi = %s' % pool2xi)
    
    #
    # Tests
    #
    TASKS = [(mul, (i, 7)) for i in range(10)] + \
            [(plus, (i, 8)) for i in range(10)]

    results = [pool2xi.apply_async(calculate, t) for t in TASKS]
    imap_it = pool2xi.imap(calculatestar, TASKS)
    imap_unordered_it = pool2xi.imap_unordered(calculatestar, TASKS)
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (01) Ordered results using pool2xi.apply_async():')
    for r in results:
        print ('\t', r.get())
    print()
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (02) Ordered results using pool2xi.imap():')    
    for x in imap_it:
        print ('\t', x)
    print()
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (03) Unordered results using pool2xi.imap_unordered():')
    for x in imap_unordered_it:
        print ('\t', x)
    print()
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (04) Ordered results using pool2xi.map() --- will block till complete:')
    for x in pool2xi.map(calculatestar, TASKS):
        print ('\t', x)
    print()

    #
    # Simple benchmarks
    #
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    N = 100000
    print ('\n DragonXi Test (05) def pow3(x): return x**3')

    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    t = time.time()
    A = map(pow3, range(N))
    print ('\n DragonXi Test (06) \tmap(pow3, range(%d)):\n\t\t%s seconds' % \
          (N, time.time() - t))
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    t = time.time()
    B = pool2xi.map(pow3, range(N))
    print ('\n DragonXi Test (07) \tpool.map(pow3, range(%d)):\n\t\t%s seconds' % \
          (N, time.time() - t))
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    t = time.time()
    C = list(pool2xi.imap(pow3, range(N), chunksize=N//8))
    print ('\n DragonXi Test (08) \tlist(pool2xi.imap(pow3, range(%d), chunksize=%d)):\n\t\t%s' \
          ' seconds' % (N, N//8, time.time() - t))

    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ()
    L = [None] * 1000000
    print ('\n DragonXi Test (09) def noop(x): pass')

    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (10) L = [None] * 1000000')

    
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    t = time.time()
    A = map(noop, L)
    print ('\n DragonXi Test (11) \tmap(noop, L):\n\t\t%s seconds' % \
          (time.time() - t))

    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------          
    t = time.time()
    B = pool2xi.map(noop, L)
    print ('\n DragonXi Test (12) \tpool2xi.map(noop, L):\n\t\t%s seconds' % \
          (time.time() - t))

    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------          
    t = time.time()
    C = list(pool2xi.imap(noop, L, chunksize=len(L)//8))
    print ('\n DragonXi Test (13) \tlist(pool2xi.imap(noop, L, chunksize=%d)):\n\t\t%s seconds' % \
          (len(L)//8, time.time() - t))

    #assert (A == B == C, (len(A), len(B), len(C)))
    print ()
    del A, B, C, L

    #
    # Test error handling
    #
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (14) Testing error handling:')

    try:
        print (pool2xi.apply(f, (5,)))
    except ZeroDivisionError:
        print ('\n Got ZeroDivisionError as expected from pool.apply()')
    else:
        #raise AssertionError, 'expected ZeroDivisionError'
        raise AssertionError('expected ZeroDivisionError')
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------    
    try:
        print ('\n DragonXi Test (15) ', pool2xi.map(f, range(10)))
    except ZeroDivisionError:
        print ('\tGot ZeroDivisionError as expected from pool2xi.map()')
    else:
        raise AssertionError ('expected ZeroDivisionError')
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    try:
        print ('\n DragonXi Test (16) ', list(pool2xi.imap(f, range(10))))
    except ZeroDivisionError:
        print ('\n Got ZeroDivisionError as expected from list(pool2xi.imap())')
    else:
        raise AssertionError ('expected ZeroDivisionError')

    it = pool2xi.imap(f, range(10))
    for i in range(10):
        try:
            x = it.next()
        except ZeroDivisionError:
            if i == 5:
                pass
        except StopIteration:
            break
        else:
            if i == 5:
                raise AssertionError ('expected ZeroDivisionError')
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    assert (i == 9)
    print ('\n DragonXi Test (17) \tGot ZeroDivisionError as expected from IMapIterator.next()')
    print ()
    #
    # Testing timeouts
    #
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (18) Testing ApplyResult.get() with timeout:',)
    res = pool2xi.apply_async(calculate, TASKS[0])
    while 1:
        sys.stdout.flush()
        try:
            sys.stdout.write('\n\t%s' % res.get(0.02))
            break
        except multiprocessing.TimeoutError:
            sys.stdout.write('.')
    print()

    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (19) Testing IMapIterator.next() with timeout:')
    it = pool2xi.imap(calculatestar, TASKS)
    while 1:
        sys.stdout.flush()
        try:
            sys.stdout.write('\n\t%s' % it.next(0.02))
        except StopIteration:
            break
        except multiprocessing.TimeoutError:
            sys.stdout.write('.')
    print()


    #
    # Testing callback
    #
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (20) Testing callback:')

    A = []
    B = [56, 0, 1, 8, 27, 64, 125, 216, 343, 512, 729]

    r = pool2xi.apply_async(mul, (7, 8), callback=A.append)
    r.wait()

    r = pool2xi.map_async(pow3, range(10), callback=A.extend)
    r.wait()

    if A == B:
        print ('\tcallbacks succeeded\n')
    else:
        print ('\t*** callbacks failed\n\t\t%s != %s\n' % (A, B))

    #
    # Check there are no outstanding tasks
    #

    # assert (not pool2xi._cache, 'cache = %r' % pool2xi._cache) <--- removed

    #
    # Check close() methods
    #
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (21) Testing close():')

    for worker in pool2xi._pool:
        assert (worker.is_alive())

    result = pool2xi.apply_async(time.sleep, [0.5])
    pool2xi.close()
    pool2xi.join()

    assert (result.get() is None)

    for worker in pool2xi._pool:
        assert (not worker.is_alive())

    print ('\tclose() succeeded\n')

    #
    # Check terminate() method
    #
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (22) Testing terminate()')
    pool2xi = Pool2Xi(2)
    DELTA = 0.1
    ignore = pool2xi.apply(pow3, [2])
    results = [pool2xi.apply_async(time.sleep, [DELTA]) for i in range(100)]
    pool2xi.terminate()
    pool2xi.join()
    for worker in pool2xi._pool:
        assert (not worker.is_alive())
    print ('\tterminate() succeeded\n')

    #
    # Check garbage collection
    #
    # 
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n DragonXi Test (23) Testing garbage collection')

    pool2xi = Pool2Xi(2)
    DELTA = 0.1
    processes = pool2xi._pool
    ignore = pool2xi.apply(pow3, [2])
    results = [pool2xi.apply_async(time.sleep, [DELTA]) for i in range(100)]

    results = pool2xi = None

    time.sleep(DELTA * 2)

    for worker in processes:
        assert (not worker.is_alive())

    print ('\tgarbage collection succeeded\n')
    # --------------------------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------------------------
    print ('\n All tests with dragonized multiprocessing modules done!')

        
# -------------------------------------------------------------------------------------------------
# main                               
# -------------------------------------------------------------------------------------------------
if __name__ == '__main__':



    print ('\n test_math2_mp_xi2.py: testing calculations with standard multiprocessing \n')
    test1()
    
    print ('\n test_math2_mp_xi2.py: testing calculations with dragonized multiprocessing \n')
    test2()

# -------------------------------------------------------------------------------------------------
# test2 Output with dragonized multiprocessing                                
# -------------------------------------------------------------------------------------------------
#  test_math2_mp_xi2.py: testing calculations with dragonized multiprocessing 
# 
# 
#  DragonXi Test(00) cpu_corecount2xi() = 2
# 
# Creating dragonized pool with 4 processes
# 
# pool2xi = <pool2xi.Pool2Xi object at 0x00000000033EEF60>
# 
#  DragonXi Test (01) Ordered results using pool2xi.apply_async():
# 	 PoolWorker-1 says that mul(0, 7) = 0
# 	 PoolWorker-3 says that mul(1, 7) = 7
# 	 PoolWorker-4 says that mul(2, 7) = 14
# 	 PoolWorker-2 says that mul(3, 7) = 21
# 	 PoolWorker-3 says that mul(4, 7) = 28
# 	 PoolWorker-1 says that mul(5, 7) = 35
# 	 PoolWorker-1 says that mul(6, 7) = 42
# 	 PoolWorker-1 says that mul(7, 7) = 49
# 	 PoolWorker-4 says that mul(8, 7) = 56
# 	 PoolWorker-3 says that mul(9, 7) = 63
# 	 PoolWorker-2 says that plus(0, 8) = 8
# 	 PoolWorker-4 says that plus(1, 8) = 9
# 	 PoolWorker-1 says that plus(2, 8) = 10
# 	 PoolWorker-3 says that plus(3, 8) = 11
# 	 PoolWorker-2 says that plus(4, 8) = 12
# 	 PoolWorker-4 says that plus(5, 8) = 13
# 	 PoolWorker-2 says that plus(6, 8) = 14
# 	 PoolWorker-1 says that plus(7, 8) = 15
# 	 PoolWorker-1 says that plus(8, 8) = 16
# 	 PoolWorker-3 says that plus(9, 8) = 17
# 
# 
#  DragonXi Test (02) Ordered results using pool2xi.imap():
# 	 PoolWorker-4 says that mul(0, 7) = 0
# 	 PoolWorker-2 says that mul(1, 7) = 7
# 	 PoolWorker-3 says that mul(2, 7) = 14
# 	 PoolWorker-2 says that mul(3, 7) = 21
# 	 PoolWorker-1 says that mul(4, 7) = 28
# 	 PoolWorker-2 says that mul(5, 7) = 35
# 	 PoolWorker-3 says that mul(6, 7) = 42
# 	 PoolWorker-4 says that mul(7, 7) = 49
# 	 PoolWorker-4 says that mul(8, 7) = 56
# 	 PoolWorker-2 says that mul(9, 7) = 63
# 	 PoolWorker-1 says that plus(0, 8) = 8
# 	 PoolWorker-4 says that plus(1, 8) = 9
# 	 PoolWorker-3 says that plus(2, 8) = 10
# 	 PoolWorker-2 says that plus(3, 8) = 11
# 	 PoolWorker-3 says that plus(4, 8) = 12
# 	 PoolWorker-2 says that plus(5, 8) = 13
# 	 PoolWorker-1 says that plus(6, 8) = 14
# 	 PoolWorker-2 says that plus(7, 8) = 15
# 	 PoolWorker-2 says that plus(8, 8) = 16
# 	 PoolWorker-4 says that plus(9, 8) = 17
# 
# 
#  DragonXi Test (03) Unordered results using pool2xi.imap_unordered():
# 	 PoolWorker-2 says that mul(0, 7) = 0
# 	 PoolWorker-3 says that mul(2, 7) = 14
# 	 PoolWorker-1 says that mul(1, 7) = 7
# 	 PoolWorker-2 says that mul(3, 7) = 21
# 	 PoolWorker-4 says that mul(7, 7) = 49
# 	 PoolWorker-1 says that mul(5, 7) = 35
# 	 PoolWorker-3 says that mul(4, 7) = 28
# 	 PoolWorker-2 says that mul(6, 7) = 42
# 	 PoolWorker-1 says that mul(9, 7) = 63
# 	 PoolWorker-3 says that plus(0, 8) = 8
# 	 PoolWorker-4 says that mul(8, 7) = 56
# 	 PoolWorker-2 says that plus(1, 8) = 9
# 	 PoolWorker-3 says that plus(3, 8) = 11
# 	 PoolWorker-4 says that plus(4, 8) = 12
# 	 PoolWorker-1 says that plus(2, 8) = 10
# 	 PoolWorker-2 says that plus(5, 8) = 13
# 	 PoolWorker-4 says that plus(7, 8) = 15
# 	 PoolWorker-2 says that plus(9, 8) = 17
# 	 PoolWorker-3 says that plus(6, 8) = 14
# 	 PoolWorker-1 says that plus(8, 8) = 16
# 
# 
#  DragonXi Test (04) Ordered results using pool2xi.map() --- will block till complete:
# 	 PoolWorker-4 says that mul(0, 7) = 0
# 	 PoolWorker-4 says that mul(1, 7) = 7
# 	 PoolWorker-2 says that mul(2, 7) = 14
# 	 PoolWorker-2 says that mul(3, 7) = 21
# 	 PoolWorker-3 says that mul(4, 7) = 28
# 	 PoolWorker-3 says that mul(5, 7) = 35
# 	 PoolWorker-1 says that mul(6, 7) = 42
# 	 PoolWorker-1 says that mul(7, 7) = 49
# 	 PoolWorker-1 says that mul(8, 7) = 56
# 	 PoolWorker-1 says that mul(9, 7) = 63
# 	 PoolWorker-3 says that plus(0, 8) = 8
# 	 PoolWorker-3 says that plus(1, 8) = 9
# 	 PoolWorker-1 says that plus(2, 8) = 10
# 	 PoolWorker-1 says that plus(3, 8) = 11
# 	 PoolWorker-4 says that plus(4, 8) = 12
# 	 PoolWorker-4 says that plus(5, 8) = 13
# 	 PoolWorker-2 says that plus(6, 8) = 14
# 	 PoolWorker-2 says that plus(7, 8) = 15
# 	 PoolWorker-4 says that plus(8, 8) = 16
# 	 PoolWorker-4 says that plus(9, 8) = 17
# 
# 
#  DragonXi Test (05) def pow3(x): return x**3
# 
#  DragonXi Test (06) 	map(pow3, range(100000)):
# 		0.0 seconds
# 
#  DragonXi Test (07) 	pool.map(pow3, range(100000)):
# 		0.228999853134 seconds
# 
#  DragonXi Test (08) 	list(pool2xi.imap(pow3, range(100000), chunksize=12500)):
# 		0.173000097275 seconds
# 
# 
#  DragonXi Test (09) def noop(x): pass
# 
#  DragonXi Test (10) L = [None] * 1000000
# 
#  DragonXi Test (11) 	map(noop, L):
# 		0.0 seconds
# 
#  DragonXi Test (12) 	pool2xi.map(noop, L):
# 		0.444999933243 seconds
# 
#  DragonXi Test (13) 	list(pool2xi.imap(noop, L, chunksize=125000)):
# 		0.641999959946 seconds
# 
# 
#  DragonXi Test (14) Testing error handling:
# 
#  Got ZeroDivisionError as expected from pool.apply()
# 	Got ZeroDivisionError as expected from pool2xi.map()
# 
#  Got ZeroDivisionError as expected from list(pool2xi.imap())
# 
#  DragonXi Test (17) 	Got ZeroDivisionError as expected from IMapIterator.next()
# 
# 
#  DragonXi Test (18) Testing ApplyResult.get() with timeout:
# .
# 	PoolWorker-3 says that mul(0, 7) = 0
# 
#  DragonXi Test (19) Testing IMapIterator.next() with timeout:
# ..............
# 	PoolWorker-1 says that mul(0, 7) = 0....
# 	PoolWorker-2 says that mul(1, 7) = 7
# 	PoolWorker-4 says that mul(2, 7) = 14
# 	PoolWorker-3 says that mul(3, 7) = 21
# 	PoolWorker-4 says that mul(4, 7) = 28
# 	PoolWorker-4 says that mul(5, 7) = 35.........
# 	PoolWorker-4 says that mul(6, 7) = 42
# 	PoolWorker-1 says that mul(7, 7) = 49
# 	PoolWorker-1 says that mul(8, 7) = 56
# 	PoolWorker-3 says that mul(9, 7) = 63
# 	PoolWorker-2 says that plus(0, 8) = 8......
# 	PoolWorker-1 says that plus(1, 8) = 9......
# 	PoolWorker-2 says that plus(2, 8) = 10
# 	PoolWorker-3 says that plus(3, 8) = 11.......
# 	PoolWorker-4 says that plus(4, 8) = 12
# 	PoolWorker-1 says that plus(5, 8) = 13...
# 	PoolWorker-3 says that plus(6, 8) = 14...........
# 	PoolWorker-2 says that plus(7, 8) = 15
# 	PoolWorker-1 says that plus(8, 8) = 16.
# 	PoolWorker-4 says that plus(9, 8) = 17
# 
#  DragonXi Test (20) Testing callback:
# 	callbacks succeeded
# 
# 
#  DragonXi Test (21) Testing close():
# 	close() succeeded
# 
# 
#  DragonXi Test (22) Testing terminate()
# 	terminate() succeeded
# 
# 
#  DragonXi Test (23) Testing garbage collection
# 	garbage collection succeeded
# 
# 
#  All tests with dragonized multiprocessing modules done!
# -------------------------------------------------------------------------------------------------
# test1 Output with standard multiprocessing                                
# -------------------------------------------------------------------------------------------------
#  test_math2_mp_xi2.py: testing calculations with standard multiprocessing 
# 
# cpu_count() = 2
# 
# Creating pool with 4 processes
# 
# pool = <multiprocessing.pool.Pool object at 0x0000000003394C88>
# Ordered results using pool.apply_async():
# 	 PoolWorker-3 says that mul(0, 7) = 0
# 	 PoolWorker-4 says that mul(1, 7) = 7
# 	 PoolWorker-4 says that mul(2, 7) = 14
# 	 PoolWorker-1 says that mul(3, 7) = 21
# 	 PoolWorker-2 says that mul(4, 7) = 28
# 	 PoolWorker-4 says that mul(5, 7) = 35
# 	 PoolWorker-4 says that mul(6, 7) = 42
# 	 PoolWorker-3 says that mul(7, 7) = 49
# 	 PoolWorker-2 says that mul(8, 7) = 56
# 	 PoolWorker-1 says that mul(9, 7) = 63
# 	 PoolWorker-4 says that plus(0, 8) = 8
# 	 PoolWorker-2 says that plus(1, 8) = 9
# 	 PoolWorker-1 says that plus(2, 8) = 10
# 	 PoolWorker-3 says that plus(3, 8) = 11
# 	 PoolWorker-4 says that plus(4, 8) = 12
# 	 PoolWorker-1 says that plus(5, 8) = 13
# 	 PoolWorker-2 says that plus(6, 8) = 14
# 	 PoolWorker-1 says that plus(7, 8) = 15
# 	 PoolWorker-3 says that plus(8, 8) = 16
# 	 PoolWorker-4 says that plus(9, 8) = 17
# 
# Ordered results using pool.imap():
# 	 PoolWorker-1 says that mul(0, 7) = 0
# 	 PoolWorker-4 says that mul(1, 7) = 7
# 	 PoolWorker-1 says that mul(2, 7) = 14
# 	 PoolWorker-3 says that mul(3, 7) = 21
# 	 PoolWorker-2 says that mul(4, 7) = 28
# 	 PoolWorker-3 says that mul(5, 7) = 35
# 	 PoolWorker-2 says that mul(6, 7) = 42
# 	 PoolWorker-4 says that mul(7, 7) = 49
# 	 PoolWorker-3 says that mul(8, 7) = 56
# 	 PoolWorker-1 says that mul(9, 7) = 63
# 	 PoolWorker-4 says that plus(0, 8) = 8
# 	 PoolWorker-4 says that plus(1, 8) = 9
# 	 PoolWorker-3 says that plus(2, 8) = 10
# 	 PoolWorker-2 says that plus(3, 8) = 11
# 	 PoolWorker-1 says that plus(4, 8) = 12
# 	 PoolWorker-4 says that plus(5, 8) = 13
# 	 PoolWorker-2 says that plus(6, 8) = 14
# 	 PoolWorker-1 says that plus(7, 8) = 15
# 	 PoolWorker-3 says that plus(8, 8) = 16
# 	 PoolWorker-2 says that plus(9, 8) = 17
# 
# Unordered results using pool.imap_unordered():
# 	 PoolWorker-2 says that mul(1, 7) = 7
# 	 PoolWorker-2 says that mul(2, 7) = 14
# 	 PoolWorker-1 says that mul(0, 7) = 0
# 	 PoolWorker-3 says that mul(4, 7) = 28
# 	 PoolWorker-2 says that mul(5, 7) = 35
# 	 PoolWorker-2 says that mul(8, 7) = 56
# 	 PoolWorker-4 says that mul(3, 7) = 21
# 	 PoolWorker-3 says that mul(7, 7) = 49
# 	 PoolWorker-1 says that mul(6, 7) = 42
# 	 PoolWorker-4 says that plus(0, 8) = 8
# 	 PoolWorker-1 says that plus(2, 8) = 10
# 	 PoolWorker-2 says that mul(9, 7) = 63
# 	 PoolWorker-3 says that plus(1, 8) = 9
# 	 PoolWorker-4 says that plus(3, 8) = 11
# 	 PoolWorker-2 says that plus(5, 8) = 13
# 	 PoolWorker-4 says that plus(7, 8) = 15
# 	 PoolWorker-1 says that plus(4, 8) = 12
# 	 PoolWorker-4 says that plus(9, 8) = 17
# 	 PoolWorker-3 says that plus(6, 8) = 14
# 	 PoolWorker-2 says that plus(8, 8) = 16
# 
# Ordered results using pool.map() --- will block till complete:
# 	 PoolWorker-1 says that mul(0, 7) = 0
# 	 PoolWorker-1 says that mul(1, 7) = 7
# 	 PoolWorker-4 says that mul(2, 7) = 14
# 	 PoolWorker-4 says that mul(3, 7) = 21
# 	 PoolWorker-3 says that mul(4, 7) = 28
# 	 PoolWorker-3 says that mul(5, 7) = 35
# 	 PoolWorker-2 says that mul(6, 7) = 42
# 	 PoolWorker-2 says that mul(7, 7) = 49
# 	 PoolWorker-3 says that mul(8, 7) = 56
# 	 PoolWorker-3 says that mul(9, 7) = 63
# 	 PoolWorker-1 says that plus(0, 8) = 8
# 	 PoolWorker-1 says that plus(1, 8) = 9
# 	 PoolWorker-4 says that plus(2, 8) = 10
# 	 PoolWorker-4 says that plus(3, 8) = 11
# 	 PoolWorker-2 says that plus(4, 8) = 12
# 	 PoolWorker-2 says that plus(5, 8) = 13
# 	 PoolWorker-1 says that plus(6, 8) = 14
# 	 PoolWorker-1 says that plus(7, 8) = 15
# 	 PoolWorker-4 says that plus(8, 8) = 16
# 	 PoolWorker-4 says that plus(9, 8) = 17
# 
# def pow3(x): return x**3
# 	map(pow3, range(100000)):
# 		0.0 seconds
# 	pool.map(pow3, range(100000)):
# 		0.249000072479 seconds
# 	list(pool.imap(pow3, range(100000), chunksize=12500)):
# 		0.155999898911 seconds
# 
# def noop(x): pass
# L = [None] * 1000000
# 	map(noop, L):
# 		0.0 seconds
# 	pool.map(noop, L):
# 		0.46799993515 seconds
# 	list(pool.imap(noop, L, chunksize=125000)):
# 		0.608000040054 seconds
# 
# Testing error handling:
# 	Got ZeroDivisionError as expected from pool.apply()
# 	Got ZeroDivisionError as expected from pool.map()
# 	Got ZeroDivisionError as expected from list(pool.imap())
# 	Got ZeroDivisionError as expected from IMapIterator.next()
# 
# Testing ApplyResult.get() with timeout:
# .........
# 	PoolWorker-1 says that mul(0, 7) = 0
# Testing IMapIterator.next() with timeout:
# 
# 	PoolWorker-2 says that mul(0, 7) = 0
# 	PoolWorker-3 says that mul(1, 7) = 7........
# 	PoolWorker-4 says that mul(2, 7) = 14
# 	PoolWorker-1 says that mul(3, 7) = 21
# 	PoolWorker-2 says that mul(4, 7) = 28
# 	PoolWorker-3 says that mul(5, 7) = 35
# 	PoolWorker-2 says that mul(6, 7) = 42
# 	PoolWorker-1 says that mul(7, 7) = 49.........
# 	PoolWorker-2 says that mul(8, 7) = 56
# 	PoolWorker-1 says that mul(9, 7) = 63.....
# 	PoolWorker-3 says that plus(0, 8) = 8
# 	PoolWorker-4 says that plus(1, 8) = 9
# 	PoolWorker-1 says that plus(2, 8) = 10
# 	PoolWorker-1 says that plus(3, 8) = 11
# 	PoolWorker-1 says that plus(4, 8) = 12.
# 	PoolWorker-4 says that plus(5, 8) = 13
# 	PoolWorker-2 says that plus(6, 8) = 14......
# 	PoolWorker-2 says that plus(7, 8) = 15....
# 	PoolWorker-1 says that plus(8, 8) = 16
# 	PoolWorker-3 says that plus(9, 8) = 17
# Testing callback:
# 	callbacks succeeded
# 
# Testing close():
# 	close() succeeded
# 
# Testing terminate():
# 	terminate() succeeded
# 
# Testing garbage collection:
# 	garbage collection succeeded
# 
# -------------------------------------------------------------------------------------------------
# History                               
# -------------------------------------------------------------------------------------------------
# see math2_mp_xi2.py
#        
# -------------------------------------------------------------------------------------------------
# End Of test_math2_mp_xi2.py                               
# -------------------------------------------------------------------------------------------------